#!/usr/bin/env python

from bisect import bisect_left, insort_left

# Time management.

def have_conflict(freebusy, periods, get_conflicts=False):

    """
    Return whether any period in 'freebusy' overlaps with the given 'periods',
    returning a collection of such overlapping periods if 'get_conflicts' is
    set to a true value.
    """

    conflicts = []
    for start, end in periods:
        overlapping = period_overlaps(freebusy, (start, end), get_conflicts)
        if overlapping:
            if get_conflicts:
                conflicts += overlapping
            else:
                return True

    if get_conflicts:
        return conflicts
    else:
        return False

def insert_period(freebusy, period):
    insort_left(freebusy, period)

def remove_period(freebusy, uid):
    i = 0
    while i < len(freebusy):
        t = freebusy[i]
        if len(t) >= 3 and t[2] == uid:
            del freebusy[i]
        else:
            i += 1

def period_overlaps(freebusy, period, get_periods=False):

    """
    Return whether any period in 'freebusy' overlaps with the given 'period',
    returning a collection of overlapping periods if 'get_periods' is set to a
    true value.
    """

    dtstart, dtend = period[:2]
    found = bisect_left(freebusy, (dtstart, dtend, None, None))

    overlapping = []

    # Find earlier overlapping periods.

    i = found

    while i > 0 and freebusy[i - 1][1] > dtstart:
        if get_periods:
            overlapping.insert(0, freebusy[i - 1])
        else:
            return True
        i -= 1

    # Find later overlapping periods.

    i = found

    while i < len(freebusy) and (dtend is None or freebusy[i][0] < dtend):
        if get_periods:
            overlapping.append(freebusy[i])
        else:
            return True
        i += 1

    if get_periods:
        return overlapping
    else:
        return False

# Period layout.

def get_scale(l):

    """
    Return an ordered time scale from the given list 'l' of tuples, with the
    first two elements of each tuple being start and end times.
    """

    scale = {}

    for t in l:
        start, end = t[:2]

        # Add a point and this event to the starting list.

        if not scale.has_key(start):
            scale[start] = [], []
        scale[start][0].append(t)

        # Add a point and this event to the ending list.

        if not scale.has_key(end):
            scale[end] = [], []
        scale[end][1].append(t)

    scale = scale.items()
    scale.sort()
    return scale

def get_slots(l):

    """
    Return an ordered list of time slots from the given list 'l' of tuples, with
    the first two elements of each tuple being start and end times.

    Each slot is a tuple containing a point in time for the start of the slot,
    together with a list of parallel event tuples, each tuple containing the
    original details of an event.
    """

    slots = []
    active = []

    for point, (starting, ending) in get_scale(l):

        # Discard all active events ending at or before this start time.

        for t in ending:
            i = active.index(t)
            active[i] = None

        for t in starting:
            try:
                i = active.index(None)
                active[i] = t
            except ValueError:
                active.append(t)

        while active and active[-1] is None:
            active.pop()

        slots.append((point, active[:]))

    return slots

def get_spans(slots):

    "Inspect the given 'slots', returning a mapping of event uids to spans."

    points = [point for point, active in slots]
    spans = {}

    for point, active in slots:
        for t in active:
            if t:
                start, end, uid = t[:3]
                start_slot = points.index(start)
                end_slot = points.index(end)
                spans[uid] = end_slot - start_slot

    return spans

# vim: tabstop=4 expandtab shiftwidth=4
